"""
All of the utilities that are used in writing tests for Peeves
"""

from .Timer import Timer
import unittest, os, sys, argparse

__all__ = [
    "TestRunner",
    "DebugTests",
    "ValidationTests",
    "TimingTests",
    "TestManager",
    "DataGenerator",
    "load_tests",
    "validationTest",
    "debugTest",
    "dataGenTest",
    "timeitTest",
    "timingTest",
    "inactiveTest"
]

class TestManager:
    """
    Manages the run of the tests
    """

    log_file = "test_results.txt"
    log_results = False
    quiet_mode = False
    debug_tests = True
    validation_tests = False
    timing_tests = False
    data_gen_tests = False
    test_files = "All"
    test_name = None
    def __init__(self,
                 test_root=None, test_dir=None, test_data=None,
                 base_dir=None, start_dir=None,
                 test_pkg=None, test_data_ext="TestData",
                 log_file=None,
                 log_results=None,
                 quiet_mode=None,
                 debug_tests=None,
                 validation_tests=None,
                 timing_tests=None,
                 data_gen_tests=None,
                 test_files=None,
                 test_name=None
                 ):
        """
        :param test_root: the root package
        :type test_root:
        :param test_dir: the directory to load tests from (usually test_root/test_pkg)
        :type test_dir:
        :param test_data: the directory to load test data from (usually test_dir/test_data_ext)
        :type test_data:
        :param base_dir: the overall base directory to do imports from
        :type base_dir:
        :param start_dir: the directory to start test discovery from
        :type start_dir:
        :param test_pkg: the name of the python package that holds all the tests
        :type test_pkg:
        :param test_data_ext: the extension from test_dir to look for data in (usually TestData)
        :type test_data_ext:
        """
        self._base_dir = base_dir
        self._start_dir = start_dir
        self._test_root = test_root
        self._base_dir_use_default = base_dir is None
        self._test_dir = test_dir
        self._test_dir_use_default = test_dir is None
        self._test_data = test_data
        self._test_data_use_default = test_data is None
        self._test_pkg = test_pkg
        self._test_pkg_validated = False
        self.data_ext = test_data_ext

        if log_file is not None:
            self.log_file = log_file
        if log_results is not None:
            self.log_results = log_results
        if quiet_mode is not None:
            self.quiet_mode = quiet_mode
        if debug_tests is not None:
            self.debug_tests = debug_tests
        if validation_tests is not None:
            self.validation_tests = validation_tests
        if timing_tests is not None:
            self.timing_tests = timing_tests
        if data_gen_tests is not None:
            self.data_gen_tests = data_gen_tests
        if test_files is not None:
            self.test_files = test_files
        if test_name is not None:
            self.test_name = test_name
    @property
    def test_root(self):
        if self._test_root is None:
            try:
                test_root = [ a for a in sys.argv if os.path.isdir(a) ][0] # you can pass the directory to run the tests as the first sys.argv arg
            except IndexError:
                test_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__))) # or we'll assume it's two dirs up from here
            sys.path.insert(0, test_root) # not sure exactly what this does... might want to be more targeted with it
            self._test_root = test_root
        return self._test_root
    @test_root.setter
    def test_root(self, root):
        self._test_root = root
        if self._base_dir_use_default:
            self._base_dir = None
        if self._test_dir_use_default:
            self._test_dir = None
        if self._test_data_use_default:
            self._test_data = None
    @property
    def base_dir(self):
        if self._base_dir is None:
            self._base_dir = self.test_root
        return self._base_dir
    @base_dir.setter
    def base_dir(self, d):
        self._base_dir = d
        if d is not None:
            self._base_dir_use_default = False
    @property
    def start_dir(self):
        if self._start_dir is None:
            return self.test_dir
        return self._start_dir
    @start_dir.setter
    def start_dir(self, d):
        self._start_dir = d
        if d is not None:
            self._start_dir_use_default = False
    @property
    def test_pkg(self):
        if not self._test_pkg_validated:
            root = self.test_root
            # TODO: find some way to check it to figure out how many . we need to go up...
            # for now we'll just leave it, though
            if self._test_pkg is None:
                self._test_pkg = "Tests"
            if "." not in self._test_pkg:
                self._test_pkg = "."*(len(__package__.split(".")) - 1) + self._test_pkg
                # a basic guess as to what'll get us to the right spot...
            self._test_pkg_validated = True
        return self._test_pkg
    @test_pkg.setter
    def test_pkg(self, pkg):
        self._test_pkg = pkg
        self._test_pkg_validated = False
    @property
    def test_dir(self):
        # the Tests package _must_ be in the parent repository
        if self._test_dir is None:
            self._test_dir = os.path.join(self.test_root, self.test_pkg.split(".")[-1])
            if not os.path.isdir(self._test_dir) and self.test_pkg[0] == ".":
                raise IOError(
                    "Peeves expects a '{}' package at {} to hold all the tests because I wrote it bad".format(
                        self.test_pkg,
                        self.test_root
                        )
                    )
        return self._test_dir
    @test_dir.setter
    def test_dir(self, d):
        self._test_dir = d
        if d is not None:
            self._test_dir_use_default = False
    @property
    def test_data_dir(self):
        if self._test_data is None:
            self._test_data = os.path.join(self.test_dir, self.data_ext)
        return self._test_data
    @test_data_dir.setter
    def test_data_dir(self, d):
        self._test_data = d
        if d is not None:
            self._test_data_use_default = False
    @classmethod
    def test_data(cls, filename):
        return os.path.join(cls().test_data_dir, filename)

    @classmethod
    def collect_run_args(cls):
        parser = argparse.ArgumentParser(description='Process test configurations')

        def parse_bool(s):
            if s in {"False", '0'}:
                return False
            else:
                return True

        parser.add_argument('-q', dest='quiet', metavar='-q',
                            type=bool, nargs='?', default=cls.quiet_mode,
                            help='whether to run the tests in quiet mode')
        parser.add_argument('-d', dest='debug', metavar='-d',
                            type=parse_bool, nargs='?', default=cls.debug_tests,
                            help='whether to run the debug tests')
        parser.add_argument('-v', dest='validate', metavar='-v',
                            type=parse_bool, nargs='?', default=cls.validation_tests,
                            help='whether to run the validation tests')
        parser.add_argument('-t', dest='timing', metavar='-t',
                            type=parse_bool, nargs='?', default=cls.timing_tests,
                            help='whether to run the timing tests')
        parser.add_argument('-g', dest='data_gen', metavar='-g',
                            type=parse_bool, nargs='?', default=cls.data_gen_tests,
                            help='whether to run the data generating tests')
        parser.add_argument('-l', dest='log', metavar='-l',
                            type=parse_bool, nargs='?', default=cls.log_results,
                            help='whether to log results')
        parser.add_argument('-L', dest='logfile', metavar='-L',
                            type=str, nargs='?', default=cls.log_file,
                            help='whether to log results')
        parser.add_argument('-f', dest='testfile', metavar='-f',
                            type=str, nargs='?', default=cls.test_files,
                            help='which tests to run')
        parser.add_argument('-n', dest='testname', metavar='-n',
                            type=str, nargs='?', default="",
                            help='name of specific test to run')
        args = parser.parse_args()

        return dict(
            quiet_mode=True if args.quiet is None else args.quiet,
            debug_tests=True if args.debug is None else args.debug,
            validation_tests=True if args.validate is None else args.validate,
            timing_tests=True if args.timing is None else args.timing,
            data_gen_tests=True if args.data_gen is None else args.data_gen,
            log_results=True if args.log is None else args.log,
            log_file=args.logfile,
            test_name=None if args.testname.lower().strip() == "" else args.testname,
            test_files='All' if args.testfile.lower() == "all" else args.testfile.split(",")
        )

    def get_log_file(self, log_results=None, log_file=None, syserr_redirect=True):
        log_results = self.log_results if log_results is None else log_results
        log_file = self.log_file if log_file is None else log_file
        if os.path.abspath(log_file) != log_file:
            log_file = os.path.join(self.test_dir, log_file)
        log_stream = self._stderr_wrap(open(log_file, "w") if log_results else sys.stderr, syserr_redirect=syserr_redirect)
        return log_stream
    class _stderr_wrap:
        def __init__(self, stream, syserr_redirect=True):
            self.stream = stream
            self.syserr_redirect = syserr_redirect
            self.stderr1 = None
            self.stdout1 = None
        def __enter__(self):
            self.stderr1 = sys.stderr
            sys.stderr = self.stream
            if self.syserr_redirect:
                self.stdout1 = sys.stdout
                sys.stdout = sys.stderr
            if hasattr(self.stream, '__enter__'):
                return self.stream.__enter__()
            else:
                return self.stream
        def __exit__(self, exc_type, exc_val, exc_tb):
            sys.stderr = self.stderr1
            self.stderr1 = None
            if self.syserr_redirect:
                sys.stdout = self.stdout1
                sys.stdout1 = None
            if hasattr(self.stream, '__exit__'):
                self.stream.__exit__(exc_type, exc_val, exc_tb)

    def run_tests(self, tag, test_set, runner, log_stream):
        print(
            "\n" + "-" * 70 + "\n" + "-" * 70 + "\n" +
            "Running {} Tests:\n".format(tag),
            file=log_stream
        )
        return runner.run(test_set)

    def get_test_types(self, debug=None, validate=None, timing=None, data_gen=None):
        tests = {}
        dt = DebugTests() # drain the stack even if unused
        if debug or (debug is None and self.debug_tests):
            tests["Debug"] = dt
        vt = ValidationTests()
        if validate or (validate is None and self.validation_tests):
            tests["Validation"] = vt
        tt = TimingTests()
        if timing or (timing is None and self.timing_tests):
            tests["Timing"] = tt
        dgt = DataGenTests()
        if data_gen or (data_gen is None and self.data_gen_tests):
            tests["DataGen"] = dgt
        return tests
    def _run(self,
             log_results=None, log_file=None, quiet=None,
             syserr_redirect=True
             ):
        TestManager.test_data_dir = self.test_data_dir
        with self.get_log_file(log_results=log_results, log_file=log_file, syserr_redirect=syserr_redirect) as log_stream:
            quiet = self.quiet_mode if quiet is None else quiet
            v_level = 1 if quiet else 2
            runner = TestRunner(stream=log_stream, verbosity=v_level)

            results = [
                self.run_tests(tag, tests, runner, log_stream)
                for tag, tests in self.get_test_types().items()
                ]
        test_status = not all(res.wasSuccessful() for res in results)
        return test_status

    def load_tests(self, start_dir=None, base_dir=None):
        if start_dir is None:
            start_dir = self.start_dir
        if base_dir is None:
            base_dir = self.base_dir

        sys.path.insert(0, base_dir)
        cur = ManagedTestLoader.manager
        try:
            ManagedTestLoader.manager = self
            unittest.defaultTestLoader.discover(
                start_dir,
                top_level_dir=base_dir
            )
        finally:
            ManagedTestLoader.manager = cur
    @classmethod
    def run(cls,
            exit=True, exit_code=None,
            syserr_redirect=True,
            cmd_line=False,
            **kwargs
            ):

        if cmd_line:
            kwargs = dict(cls.collect_run_args(), **kwargs)
        manager = cls(**kwargs)
        manager.load_tests()

        test_status = manager._run(syserr_redirect=syserr_redirect)
        if exit:
            sys.exit(test_status if exit_code is None else exit_code) #should kill everything...?
        return test_status


TestCase = unittest.TestCase #just in case I want to change this up later
class DataGenerator:
    """Provides methods to generate relevant data for testing methods
    """

    seed = 15
    @classmethod
    def coords(cls, n=50):
        import numpy as np
        np.random.seed(cls.seed)
        return np.random.rand(n, 3)
    @classmethod
    def multicoords(cls, n=10, m=50):
        import numpy as np
        np.random.seed(cls.seed)
        return np.random.rand(n, m, 3)
    @classmethod
    def mats(cls, n=1):
        import numpy as np
        np.random.seed(cls.seed)
        return np.random.rand(n, 3, 3)
    @classmethod
    def vecs(cls, n=1):
        import numpy as np
        np.random.seed(cls.seed)
        return np.random.rand(n, 3)

    @classmethod
    def angles(cls, n=50, r=(0, 360), use_rad=False):
        import numpy as np
        np.random.seed(cls.seed)
        angles = np.random.uniform(*r, size=(n,))
        if use_rad:
            angles = np.rad2deg(angles)
        return angles
    @classmethod
    def dists(cls, n=50, minmax=(.5, 1.5)):
        import numpy as np
        np.random.seed(cls.seed)
        return np.random.uniform(*minmax, size=(n,))
    @classmethod
    def zmat(cls, ncoords=15, use_rad=False):
        import numpy as np
        np.random.seed(cls.seed)
        refs1 = np.sort(np.random.randint(0, ncoords, ncoords))
        refs2 = np.sort(np.random.randint(0, ncoords, ncoords))
        refs3 = np.sort(np.random.randint(0, ncoords, ncoords))
        ass = np.arange(0, ncoords)
        refs1 = np.amin(np.array((refs1, ass)), axis=0)
        refs2 = np.amin(np.array((refs2, ass)), axis=0)
        for i,rs in enumerate(zip(refs1, refs2)):
            r1, r2 = rs
            if i > 0 and r1 == r2:
                while r1 == r2:
                    r2 = (r2 + 1) % (i + 1)
                    # print(r1, r2, i)
                refs2[i] = r2
        refs3 = np.amin(np.array((refs3, ass)), axis=0)
        for i,rs in enumerate(zip(refs1, refs2, refs3)):
            r1, r2, r3 = rs
            if i > 1 and (r1 == r3 or r2 == r3):
                while (r1 == r3 or r2 == r3):
                    r3 = (r3 + 1) % (i + 1)
                refs3[i] = r3

        # raise Exception(np.array((refs1, refs1-refs2, refs1-refs3, refs2-refs3)))
        dists = DataGenerator.dists(ncoords)
        angles = DataGenerator.angles(ncoords, (0, 180), use_rad=use_rad)
        dihedrals = DataGenerator.angles(ncoords, (0, 360), use_rad=use_rad)

        return np.array([refs1+1, dists, refs2+1, angles, refs3+1, dihedrals ]).T
    @classmethod
    def zmats(cls, m=10, ncoords=15, use_rad=False):
        import numpy as np
        np.random.seed(cls.seed)
        return np.array([DataGenerator.zmat(ncoords, use_rad) for i in range(m)])

class ManagedTestSuite(unittest.TestSuite):
    """
    A funky subclass of `TestSuite`
    that drains a stack of tests on initialization so that
    tests can be added to a global queue that is exhausted
    when called
    """
    stack = None
    def __init__(self):
        super().__init__(self.stack)
        type(self).stack = []
    @classmethod
    def queueTest(cls, test):
        cls.stack.append(test)

class DebugTests(ManagedTestSuite):
    """The set of fast tests in the test suite"""
    stack = []
class ValidationTests(ManagedTestSuite):
    """The set of slow tests in the test suite"""
    stack = []
class TimingTests(ManagedTestSuite):
    """The set of timing tests in the test suite"""
    stack = []
class InactiveTests(ManagedTestSuite):
    """The set of inactive tests in the test suite"""
    stack = []
class DataGenTests(ManagedTestSuite):
    """The set of tests in the test suite that exist to generate artefacts"""
    stack = []

def timingTest(fn):
    """
    A decorator that sets up a test to be added to the set of timing tests
    """
    timer = Timer()(fn)
    def Timing(*args, **kwargs):
        return timer(*args, **kwargs)
    return Timing
def timeitTest(**kwargs):
    """
    A decorator that sets up a test to be added to the set of timing tests
    using `timeit`
    """
    timer = Timer(**kwargs)
    def wrap(fn):
        inner_fn = timer(fn)
        def Timing(*args, **kwargs):
            return inner_fn(*args, **kwargs)
        return Timing
    return wrap

def inactiveTest(fn):
    """
    A decorator that sets up a test to be added to the set of inactive tests
    """
    def Inactive(*args, **kwargs):
        return fn(*args, **kwargs)
    Inactive.__og_name__ = fn.__name__
    return Inactive

def debugTest(fn):
    """
    A decorator that sets up a test to be added to the set of debug tests
    """
    def Debug(*args, **kwargs):
        return fn(*args, **kwargs)
    Debug.__og_name__ = fn.__name__
    return Debug

def dataGenTest(fn):
    """
    A decorator that sets up a test to be added to the set of data generation tests
    """
    def DataGen(*args, **kwargs):
        return fn(*args, **kwargs)
    DataGen.__og_name__ = fn.__name__
    return DataGen

def validationTest(fn):
    """
    A decorator that sets up a test to be added to the set of validation tests
    """
    def Validation(*args, **kwargs):
        return fn(*args, **kwargs)
    Validation.__og_name__ = fn.__name__
    return Validation

def TestRunner(**kw):
    if not "resultclass" in kw:
        kw["resultclass"] = unittest.TextTestResult
    if not "verbosity" in kw:
        kw["verbosity"] = 2
    return unittest.TextTestRunner(**kw)

_test_loader_map = {
    "Debug" : DebugTests,
    "Validation": ValidationTests,
    "Timing" : TimingTests,
    "Inactive" : InactiveTests,
    "DataGen": DataGenTests
}
class ManagedTestLoader:
    manager = None
    @classmethod
    def load_tests(cls, loader, tests, pattern):
        from itertools import chain

        pkgs = cls.manager.test_files
        names = cls.manager.test_name
        if isinstance(names, str):
            names = names.split(",")

        test_packages = None if pkgs == "All" else set(pkgs)
        if test_packages is None:
            tests = list(chain(*((t for t in suite) for suite in tests)))
        else:
            def _get_suite_name(suite):
                for test in suite:
                    return type(test).__module__.split(".")[-1]
            tests_named = {_get_suite_name(suite):suite for suite in tests}
            tests = []
            for k in tests_named:
                if k in test_packages:
                    tests.extend(tests_named[k])

        for test in tests:
            method = getattr(test, test._testMethodName)
            ttt = method.__name__
            try:
                og = method.__og_name__
            except AttributeError:
                og = ttt
            og = og.split("test_")[-1]

            if names is not None:
                if og in names:
                    for suite in _test_loader_map.values():
                        suite.queueTest(test)
            else:
                if ttt not in _test_loader_map:
                    ttt = "Debug"
                suite = _test_loader_map[ttt]
                suite.queueTest(test)
        #
        # return _test_loader_map.values()

load_tests = ManagedTestLoader.load_tests



"""
A simple documentation framework, similar to Sphinx.
It builds off of a template engine (provided by the `TemplateEngine` package) and an object-tree traversal
class (`TemplateWalker`) to fill out templates for the different types of python objects.
These templates are exported as Markdown so that they can be turned into a final website by a more sophisticated
Markdown -> HTML static site generator, like Jekyll.
Templates are extremely customizable (down to the object-level) with a flexible and powerful template mini-language
that builds off of python's built-in AST module and syntax.
Examples are also possible to provide for individual objects/modules and can also be harvested automatically from
unit tests if provided.

:long_description:
The templates provided are for modules, classes, methods, functions, and generic objects.
The mini-language used extends standard python string formatting syntax but allows for the
evaluation of a whitelisted set of commands within templates.
A full HTML/Bootstrap generator is included to allow for total customization of the generated
Markdown.
See `TemplateOps`, `TemplateFormatDirective`, `MarkdownOps`, and `MarkdownFormatDirective` for more info.
"""

__all__ = []
from .DocsBuilder import *; from .DocsBuilder import __all__ as exposed
__all__ += exposed
from .DocWalker import *; from .DocWalker import __all__ as exposed
__all__ += exposed
from .ExamplesParser import *; from .ExamplesParser import __all__ as exposed
__all__ += exposed
from .TemplateEngine import *; from .TemplateEngine import __all__ as exposed
__all__ += exposed


"""
Provides access to the JHTML interface for building quality Jupyter interfaces
"""

__all__ = []
from .HTML import *; from .HTML import __all__ as exposed
__all__ += exposed
from .Bootstrap import *; from .Bootstrap import __all__ as exposed
__all__ += exposed


"""
Implements a set of writer classes that document python objects
"""
import abc, os, sys, inspect, re, importlib, types
from collections import deque, defaultdict
from .TemplateEngine import ModuleTemplateHandler, ClassTemplateHandler
from .ExamplesParser import TestExamplesFormatter, ExamplesParser

__all__ = [
    "ModuleWriter",
    "ClassWriter",
    "FunctionWriter",
    "MethodWriter",
    "ObjectWriter",
    "IndexWriter"
]



class DocWriter(TemplateH):
    """
    A general writer class that writes a file based off a template and filling in object template specs

    :details: `DocWriter` objects are intended to do two things
     1. they manage the parsing logic to extract documentable parameters from objects
     2. they manage the process of loading the appropriate template and inserting the parameters
    This double-duty nature is likely to change in a future version of the package, being delegated to two
    subobjects that the writer then uses
    """

    template = "No template ;_;"
    template_root = "templates"
    template_name = ""
    default_template_base = os.path.dirname(__file__)
    examples_header = "## Examples"
    default_examples_root = "examples"
    default_tests_root = "tests"
    _template_cache = {}
    details_header = "## Details"
    preformat_field_handlers = {
        'examples': lambda ex,self=None:self.collapse_wrap_if_repo(self.examples_header, ex, include_opener=False) if (ex is not None and len(ex) > 0) else "",
        'details': lambda ex,self=None:self.collapse_wrap_if_repo(self.details_header, ex, open=False) if (ex is not None and len(ex) > 0) else "",
        'related': lambda objs,self=None:self.format_related_blob(objs)
    }
    protected_fields = {'id'}
    default_fields = {'details':"", "related":[]}
    def __init__(self,
                 obj,
                 out_file,
                 tree=None,

                 name=None,
                 parent=None,
                 spec=None, # extra parameters that can be used to get special behavior

                 template_directory=None,
                 examples_directory=None,
                 parent_tests=None,

                 template=None,
                 root=None,
                 ignore_paths=None,
                 examples=None,
                 tests=None,
                 formatter=None,
                 include_line_numbers=True,
                 include_link_bars=True,

                 extra_fields=None,
                 preformat_field_handlers=None,
                 ignore_missing=False,
                 strip_undocumented=False,
                 **ext
                 ):
        """
        :param obj: object to write
        :type obj:
        :param out_file: file to write to
        :type out_file: str | None
        :param tree: tree of written docs for looking stuff up
        :type tree:
        :param name: name to be used in the docs
        :type name:
        :param parent: parent object for docs purposes
        :type parent:
        :param spec: extra parameters that are usually inherited from the parent writer
        :type spec: dict | None
        :param template: template string to use when generating files
        :type template: str | None
        :param root: root directory to build off of
        :type root: str | None
        :param ignore_paths: paths to never write
        :type ignore_paths: Iterable[str]
        :param examples: path to examples to load
        :type examples: str | None
        :param tests: path to tests to load
        :type tests: str | None
        :param formatter: object that can format the stuff that Markdown supports
        :type formatter:
        """

        self.obj = obj
        self._id = None
        self._name = name
        self._parent = parent
        self._pobj = None
        self._chobj = None
        if extra_fields is None:
            extra_fields = {}
        self.extra_fields = dict(dict(self.default_fields, **ext), **extra_fields)

        self.tree = tree

        if out_file is None:
            out_file = sys.stdout
        elif isinstance(out_file, str) and os.path.isdir(out_file):
            if root is None:
                root = out_file
            out_file = os.path.join(root, *self.identifier.split("."))+".md"
        self.ignore_paths = ignore_paths if ignore_paths is not None else set()

        self.spec = {} if spec is None else spec
        self.extra_fields = dict(self.spec, **self.extra_fields)
        for k in self.extra_fields.keys() & set(self.protected_fields):
            del self.extra_fields[k]

        self.target = out_file
        if root is None:
            root = os.path.dirname(self.target)
        self.root = root

        self.fallback_template_root = 'repo_templates' if 'gh_repo' in self.extra_fields else 'templates'
        self.default_template_dir = os.path.join(self.default_template_base, self.fallback_template_root)
        self._templ_directory = (
                                    template_directory if isinstance(template_directory, str)
                                                          and os.path.isdir(template_directory) else None
        )
        self._exmpl_directory = examples_directory if isinstance(examples_directory, str) and os.path.isdir(examples_directory) else None

        self.template = self.find_template(template)
        self.examples_root = self.default_examples_root if examples is None else examples
        self.tests_root = self.default_tests_root if tests is None else tests
        self._tests = None
        self.include_line_numbers = include_line_numbers
        self.parent_tests = parent_tests

        self.include_link_bars = include_link_bars
        if preformat_field_handlers is not None:
            self.preformat_field_handlers = dict(self.preformat_field_handlers, **preformat_field_handlers)
        self.ignore_missing = ignore_missing
        self.strip_undocumented = strip_undocumented
        if strip_undocumented:
            raise NotImplementedError("currently no support for ignoring undocumented objects")
        self.formatter = MarkdownFormatter() if formatter is None else formatter

    def _clean_doc(self, doc):
        """
        Originally did a bunch of work. Now just an alias for `inspect.cleandoc`

        :param doc: a docstring
        :type doc: str
        :return: a cleaned docstring
        :rtype: str
        """
        return inspect.cleandoc(doc)

    @property
    def package_path(self):
        return self.get_package_and_url()

    @classmethod
    def load_template(cls, file):
        """
        Loads the documentation template
        for the object being documented

        :param file:
        :type file:
        :return:
        :rtype:
        """
        with open(file) as f:
            return f.read()

    def get_lineno(self):
        try:
            lineno = 1+inspect.findsource(self.obj)[1] if self.include_line_numbers else ""
        except:
            lineno = ""
        return lineno

    def resource_dir(self, spec_key, base_root):
        """
        Returns the directory for a given resource (e.g. examples or tests)
        by trying a number of different possible locations

        :param spec_key:
        :type spec_key:
        :param base_root:
        :type base_root:
        :return:
        :rtype:
        """
        if spec_key in self.spec:
            return os.path.abspath(self.spec[spec_key])
        elif os.path.isdir(os.path.abspath(base_root)):
            return base_root
        else:
            # try to inherit from the parent
            if (
                    self.tree is not None
                    and self._parent is not None
                    and self._parent in self.tree
            ):
                spec = self.tree[self._parent]
            else:
                spec = {}
            if spec_key in spec:
                return os.path.abspath(spec[spec_key])
            else:
                return os.path.join(self.root, base_root)

    def _find_template_by_name(self, name):
        test_dirs = []
        tdir = self.template_dir
        template = os.path.join(tdir, *self.identifier.split(".")) + ".md"
        if not os.path.exists(template):
            test_dirs.append(os.path.join(tdir, *self.identifier.split(".")))
            template = os.path.join(tdir, name)
        if not os.path.exists(template):
            test_dirs.append(tdir)
            def_dir = os.path.join(self.root, self.template_root)
            if not os.path.isdir(def_dir):
                def_dir = self.default_template_dir
            template = os.path.join(def_dir, name)
            if not os.path.isfile(template):
                test_dirs.append(def_dir)
                template = os.path.join(self.default_template_dir, name)
            # if os.path.isfile(template):
            #     print("no template found in {} for {}, using default".format(tdir, self.template_name))
        if os.path.exists(template):
            if template in self._template_cache:
                template = self._template_cache[template]
            else:
                tkey = template
                with open(template) as tf:
                    template = tf.read()
                self._template_cache[tkey] = template
        else:
            test_dirs.append(self.default_template_dir)
            print("no template found in {} for {}".format(test_dirs, name))
            template = self.template
        return template
    def find_template(self, template):
        """
        Finds the appropriate template for the object by looking
        in a variety of different locations

        :param template:
        :type template:
        :return:
        :rtype:
        """
        if template is None:
            template = self._find_template_by_name(self.template_name)
        elif (
            len(template.splitlines()) == 1 and
            len(os.path.split(template)) == 2
        ):
            template = self._find_template_by_name(template)

        return template

    @property
    def template_dir(self):
        if self._templ_directory is not None:
            return self._templ_directory
        elif os.path.isdir(os.path.join(self.root, self.template_root)):
            return os.path.join(self.root, self.template_root)
        else:
            return self.default_template_dir

    @property
    def examples_dir(self):
        """
        Returns the directory in which to look for examples
        :return:
        :rtype:
        """
        spec_key='examples_root'
        if self._exmpl_directory is None:
            return self.resource_dir(spec_key, self.examples_root)
        else:
            return self._exmpl_directory

    @property
    def examples_path(self):
        """
        Provides the default examples path for the object
        :return:
        :rtype:
        """
        splits = self.identifier.split(".")
        return os.path.join(*splits)+".md"
    def load_examples(self):
        """
        Loads examples for the stored object if provided
        :return:
        :rtype:
        """
        if hasattr(self.obj, '__examples__'):
            examples = os.path.join(self.examples_dir, self.obj.__examples__)
            if os.path.isfile(examples):
                with open(examples) as f:
                    return f.read()
            else:
                return self.obj.__examples__
        elif self.root is not None:
            examples = os.path.join(self.examples_dir, self.examples_path)
            if os.path.isfile(examples):
                with open(examples) as f:
                    return f.read()

    @property
    def tests_dir(self):
        """
        Returns the directory in which to look for tests
        :return:
        :rtype:
        """
        spec_key='tests_root'
        return self.resource_dir(spec_key, self.tests_root)
    @property
    def tests_path(self):
        """
        Provides the default tests path for the object
        :return:
        :rtype:
        """
        return os.path.join(*self.identifier.split(".")) + "Tests.py"
    def load_tests(self):
        """
        Loads tests for the stored object if provided
        :return:
        :rtype:
        """

        # print(">>>>", self.tests_dir, self.tests_path)
        test_str = None
        if hasattr(self.obj, '__tests__'):
            tests = os.path.join(self.tests_dir, self.obj.__tests__)
            if os.path.isfile(tests):
                with open(tests) as f:
                    test_str = f.read()
            else:
                test_str = self.obj.__tests__
        elif self.root is not None:
            tests = os.path.join(self.tests_dir, self.tests_path)
            if os.path.isfile(tests):
                with open(tests) as f:
                    test_str = f.read()
            else:
                tests = os.path.join(self.tests_dir, os.path.basename(self.tests_path))
                # print("....", tests)
                if os.path.isfile(tests):
                    with open(tests) as f:
                        test_str = f.read()

        return ExamplesParser(test_str) if test_str is not None else test_str
    @property
    def tests(self):
        if self._tests is None:
            self._tests = self.load_tests()
        return self._tests
    def get_test_markdown(self):
        tests = self.tests
        if tests is None and self.parent_tests is not None:
            tests = self.parent_tests.filter_by_name(self.name.split(".")[-1])

        formatted = TestExamplesFormatter(tests,
                                          template=self.find_template('tests.md'),
                                          example_template=self.find_template('test_example.md'),
                                          ).format() if tests is not None else ""
        # print(self.name, len(formatted))
        return formatted



    param_template = """  - `{name}`: `{type}`\n    >{description}"""
    def parse_doc(self, doc):
        """

        :param doc:
        :type doc: str
        :return:
        :rtype:
        """

        # parses a docstring based on reStructured text type specs but Markdown description
        splits = inspect.cleandoc(doc.strip()).splitlines()

        params = deque()
        param_map = {}
        extra_fields = {}

        active_tag = None
        description = []
        for line in splits:
            if line.startswith(":"):
                if line.startswith(":param"):
                    bits = line.split(":", 2)[1:]
                    name = bits[0][5:].strip()
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name":name, "type":"Any", "description":[]}
                    desc = bits[1].strip() if len(bits) == 2 else ""
                    if len(desc) > 0:
                        param_map[name]["description"].append(desc)
                elif line.startswith(":type"):
                    bits = line.split(":", 2)[1:]
                    name = bits[0][4:].strip()
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name":name, "type":"Any", "description":[]}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["type"] = t
                elif line.startswith(":return"):
                    bits = line.split(":", 2)[1:]
                    name = ":returns"
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name":name, "type":"_", "description":[]}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["description"] = t
                elif line.startswith(":rtype"):
                    bits = line.split(":", 2)[1:]
                    name = ":returns"
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name":name, "type":"_", "description":[]}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["type"] = t
                else:
                    split = line.split()[0]
                    if len(split) > 2 and split.endswith(":"):
                        bits = line.split(":", 2)[1:]
                        name = bits[0]
                        active_tag = name
                        t = bits[1].strip() if len(bits) == 2 else ""
                        if len(t) > 0:
                            extra_fields[name] = [t]
                        else:
                            extra_fields[name] = []
            else:
                if active_tag is None:
                    description.append(line.strip())
                else:
                    if active_tag in param_map:
                        param_map[active_tag]['description'].append(line)
                    else:
                        extra_fields[active_tag].append(line)
        for v in param_map.values():
            v['description'] = "\n".join(v['description'])
        for k,v in extra_fields.items():
            extra_fields[k] = "\n".join(v)

        param = []
        for p in params:
            param.append(
                self.param_template.format(**param_map[p]).strip()
            )

        return "\n".join(param), "\n".join(description), extra_fields

    def collapse_wrap_if_repo(self, header, content, name=None, open=True, include_opener=True):
        if 'gh_repo' in self.extra_fields:
            return self.formatter.format_collapse_section(header, content, name=name, open=open, include_opener=include_opener)
        else:
            return header + "\n" + content

    def resolve_object_url(self, o):
        obj = self.resolve_object(o)
        path = DocWriter.get_identifier(obj).split('.')
        return self.formatter.format_link(path[-1], "/".join(path)+".md")
    def format_related_blob(self, related):
        links = [self.resolve_object_url(o) for o in related]
        joiner = '<a>#9642;</a>'
        header = "### See Also: "
        return header + joiner.join(links)

class ModuleWriter(DocWriter):
    """A writer targeted to a module object. Just needs to write the Module metadata."""

    template_name = 'module.md'
    def __init__(self, obj, out_file, **kwargs):
        if isinstance(obj, str):
            obj = importlib.import_module(obj)
        super().__init__(obj, out_file, **kwargs)

    def get_template_params(self):
        """
        Provides module specific parameters
        :return:
        :rtype:
        """

        mod = self.obj # type: types.ModuleType
        name = mod.__name__
        ident = self.identifier
        ident_depth = len(ident.split("."))
        # get identifiers
        idents = [ DocWriter.get_identifier(getattr(mod, a)) for a in self.get_members(mod) ]
        # flattend them
        idents = [ i for i in idents if ident in i ]
        # split by qualified names
        # idents = [".".join(a.split(".")[ident_depth-1:]) for a in idents]
        # # format links
        # links = [ self.formatter.format_obj_link(l) for l in idents ]
        # if self.include_link_bars:
        #     num_cols = 3
        #     splits = []
        #     sub = []
        #     for x in links:
        #         sub.append(x)
        #         if len(sub) == num_cols:
        #             splits.append(sub)
        #             sub = []
        #     splits.append(sub + [""] * (3 - len(sub)))
        #     mems = self.formatter.format_grid_box(splits)
        # else:
        #     mems = "\n".join([ self.formatter.format_item(l) for l in links ])
        descr = mod.__doc__ if mod.__doc__ is not None else ''
        long_descr = mod.__long_doc__ if hasattr(mod, '__long_doc__') and mod.__long_doc__ is not None else ''

        ex = self.load_examples()
        tests = self.get_test_markdown()
        return {
            'id' : ident,
            'description' : descr.strip(),
            'long_description' : long_descr.strip(),
            'name': name,
            'members' : idents,
            'examples' : ex,
            'tests': tests
        }

    @classmethod
    def get_members(cls, mod):
        return (mod.__all__ if hasattr(mod, '__all__') else [])

class ClassWriter(DocWriter):
    """A writer targeted to a class"""

    template_name = 'class.md'
    def load_methods(self, function_writer=None):
        """
        Loads the methods supported by the class

        :param function_writer:
        :type function_writer:
        :return:
        :rtype:
        """

        if function_writer is None:
            function_writer = MethodWriter

        cls = self.obj
        keys = cls.__all__ if hasattr(cls, '__all__') else list(cls.__dict__.keys())

        props = []
        methods = []

        extra_fields = self.extra_fields.copy()
        pkg, file_url = self.package_path
        extra_fields['package_name'] = pkg
        extra_fields['file_url'] = file_url
        extra_fields['package_url'] = os.path.dirname(file_url)
        for k in keys:
            o = getattr(cls, k)
            if isinstance(o, (types.FunctionType, types.MethodType, classmethod, staticmethod, property)):
                if not k.startswith("_") or (k.startswith("__") and k.endswith("__")):
                    methods.append(
                        function_writer(o,
                                        tree=self.tree, parent=self.identifier, name=k,
                                        out_file=None, root=self.root,
                                        template_directory=self.template_dir,
                                        examples_directory=self.examples_dir,
                                        extra_fields=extra_fields
                                        ).format().strip()
                    )
            else:
                if not k.startswith("_"):
                    props.append(self.format_prop(k, o).strip())

        return props, methods

    def get_package_and_url(self):
        pkg, rest = self.identifier.split(".", 1)
        rest, bleh = rest.rsplit(".", 1)
        file_url = rest.replace(".", "/") + ".py"
        if 'url_base' in self.extra_fields:
            file_url = self.extra_fields['url_base'] + "/" + file_url
        # lineno = inspect.findsource(self.obj)[1]
        return pkg, file_url #+ "#L" + str(lineno) # for GitHub links

    def format_prop(self, k, o):
        return '{}: {}'.format(k, type(o).__name__)

    def get_template_params(self, function_writer = None):
        """

        :param function_writer:
        :type function_writer:
        :return:
        :rtype:
        """

        cls = self.obj # type: type
        ex = self.load_examples()
        tests = self.get_test_markdown()
        name = cls.__name__
        ident = self.identifier
        props, methods = self.load_methods(function_writer=function_writer)
        param, descr, fields = self.parse_doc(cls.__doc__ if cls.__doc__ is not None else '')
        descr = self._clean_doc(descr)
        param = self._clean_doc(param)
        if len(param) > 0:
            param = "\n" + param
        props = "\n".join(props)
        if len(props) > 0:
            props = self.formatter.format_code_block(props)+"\n"
        lineno = self.get_lineno()

        return dict({
            'id': ident,
            'name': name,
            'lineno': lineno,
            'description' : descr,
            'parameters' : param,
            'props' : props,
            'methods' : "\n\n".join(methods),
            'examples' : ex if ex is not None else "",
            'tests': tests
        }, **fields)

class FunctionWriter(DocWriter):
    """
    Writer to dump functions to file
    """

    template_name = 'function.md'
    def get_signature(self):
        return str(inspect.signature(self.obj))
    def get_template_params(self, **kwargs):
        f = self.obj # type: types.FunctionType
        ident = self.identifier
        signature = self.get_signature()
        mem_obj_pat = re.compile(" object at \w+>")
        signature = re.sub(mem_obj_pat, " instance>", signature)
        name = self.get_name()
        param, descr, fields = self.parse_doc(f.__doc__ if f.__doc__ is not None else '')
        descr = descr.strip()
        param = param.strip()
        if len(param) > 0:
            param = "\n" + param
        ex = self.load_examples()
        tests = self.get_test_markdown()
        lineno = self.get_lineno()
        return dict({
            'id': ident,
            'name' : name,
            'lineno' : lineno,
            'signature' : signature,
            'parameters' : param,
            'description' : descr,
            'examples' : ex if ex is not None else "",
            'tests': tests
        }, **fields)

    def get_package_and_url(self):
        pkg, rest = self.identifier.split(".", 1)
        rest, bleh = rest.rsplit(".", 1)
        file_url = rest.replace(".", "/") + ".py"
        if 'url_base' in self.extra_fields:
            file_url = self.extra_fields['url_base'] + "/" + file_url
        # lineno = inspect.findsource(self.obj)[1]
        return pkg, file_url #+ "#L" + str(lineno) # for GitHub links

class MethodWriter(FunctionWriter):
    """
    Writes class methods to file
    (distinct from functions since not expected to exist solo)
    """

    template_name = 'method.md'
    def get_template_params(self, **kwargs):
        params = super().get_template_params(**kwargs)
        meth = self.obj # type: types.MethodType
        decorator = ""
        if isinstance(meth, classmethod):
            decorator = 'classmethod'
        elif isinstance(meth, property):
            decorator = 'property'
        elif isinstance(meth, staticmethod):
            decorator = 'staticmethod'
        if len(decorator) > 0:
            decorator = "@" + decorator + "\n"
        params['decorator'] = decorator
        return params
    def get_signature(self):
        try:
            signature = str(inspect.signature(self.obj))
        except TypeError:  # dies on properties
            signature = "(self)"
        return signature
    @property
    def identifier(self):
        if isinstance(self.obj, property):
            return self.get_identifier(self.resolve_parent(check_tree=False)) + "." + self.get_name()
        else:
            return self.get_identifier(self.obj)

class ObjectWriter(DocWriter):
    """
    Writes general objects to file.
    Basically a fallback to support singletons and things
    of that nature.
    """

    template_name = 'object.md'
    @property
    def identifier(self):
        try:
            qualname = self.obj.__qualname__
        except AttributeError:
            qualname = self.get_identifier(type(self.obj)) + "." + self.get_name()
        qn = qualname.split(".")
        qualname = ".".join(qn[:-2] + qn[-1:]) # want to drop the class name
        # print(qualname)
        return qualname

    def check_should_write(self):
        """
        Determines whether the object really actually should be
        documented (quite permissive)
        :return:
        :rtype:
        """
        return (
                hasattr(self.obj, "__doc__")
                and hasattr(self.obj, "__name__")
                and super().check_should_write()
        )

    def get_template_params(self):

        try:
            descr = self.obj.__doc__
        except AttributeError:
            descr = "instance of "+type(self.obj).__name__

        if descr is None:
            descr = ''

        ex = self.load_examples()
        lineno = self.get_lineno()
        return {
            'id': self.identifier,
            'lineno': lineno,
            'name': self.get_name(),
            'description' : descr.strip(),
            'examples' : ex if ex is not None else ""
        }

class IndexWriter(DocWriter):
    """
    Writes an index file with all of the
    written documentation files.
    Needs some work to provide more useful info by default.
    """

    template_name = 'index.md'
    def __init__(self, *args, description=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.description = description if description is not None else "# Documentation"

    def get_identifier(cls, o):
        return 'index'

    def get_file_paths(self):
        rl = len(os.path.split(self.root))
        fs = [ "/".join(os.path.split(f)[rl-1:]) for f in self.obj ]
        return fs

    def get_template_params(self):
        files = [
            self.formatter.format_item(
                self.formatter.format_link(os.path.splitext(f.split("/")[-1])[0], f)
            ) for f in self.get_file_paths()
        ]
        return {
            'index_files' : "\n".join(files),
            'description' : self.description
        }


"""This is the test running script set up to run all tests
Peeves should be embedded in a project as a submodule at the top level
This will mean there is a "..Tests" package to load the tests from
"""

from .TestUtils import TestManager
TestManager.run(cmd_line=True)



"""Provides a little timer for testing

"""
import timeit, functools, time, sys
from collections import deque

__all__ = ["Timer"]

class Timer:

    tag_printing_times = {}
    def __init__(self, tag = None, file = sys.stderr, rounding = 5, print_times = 1, number = 1, **kw):
        self.kw = kw
        self.number = number
        self.file = file
        self.tag = tag
        self.rounding = rounding
        self.checkpoints = deque()
        self.print_times = print_times
        self.laps = []

    def get_time_list(self, time_elapsed):
        run_time = [0, 0, time_elapsed]
        if run_time[2] > 60:
            run_time[1] = int(run_time[2] / 60)
            run_time[2] = run_time[2] % 60
            if run_time[1] > 60:
                run_time[0] = int(run_time[1] / 60)
                run_time[1] = run_time[1] % 60
        return run_time

    def start(self):
        self.checkpoints.append(time.time())

    def stop(self):
        t = time.time()
        cp = self.checkpoints.pop()
        return t - cp

    def log(self):
        t = time.time()
        self.laps.append([self.checkpoints[-1], t])

    def __enter__(self):
        self.start()
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        time_elapsed = self.stop()
        self.print_timing(time_elapsed)

    def format_timing(self, time_elapsed, tag = None, steps = None):
        run_time = self.get_time_list(time_elapsed)
        if tag is None:
            tag = self.tag
        if steps is None:
            steps = self.number
        run_time_averaged = self.get_time_list(time_elapsed / steps)
        return "{0}: took {1[0]}:{1[1]}:{1[2]} per loop with {2[0]}:{2[1]}:{2[2]} overall".format(
            tag,
            [ round(x, self.rounding) for x in run_time_averaged],
            [ round(x, self.rounding) for x in run_time ]
        )

    def print_timing(self, time_elapsed, tag = None, steps = None):
        if tag is None:
            tag = self.tag
        if tag not in self.tag_printing_times:
            self.tag_printing_times[tag] = 0
        if self.tag_printing_times[tag] < self.print_times:
            print(self.format_timing(time_elapsed, tag = tag, steps = steps), file = self.file)
            self.tag_printing_times[tag] += 1

    def __call__(self, fn): # for use as a decorator
        @functools.wraps(fn)
        def timed_fn(*args, **kwargs):
            func = fn
            val = None
            time_elapsed = timeit.timeit("val = func(*args, **kwargs)", globals = locals(), **self.kw)
            self.print_timing(time_elapsed, tag = func.__name__)
            return val
        return timed_fn


"""
Peeves began life as a minor extension to the `unittest` framework.
These days it's primarily used as a documentation generator, through its `Doc` subpackage.
The documentation for Peeves itself is generated with `Peeves.Doc`
"""

from .TestUtils import *
from .Timer import *
from .Profiler import *

__all__ = []
from .TestUtils import __all__ as exposed
__all__ += exposed
from .Timer import __all__ as exposed
__all__ += exposed
from .Profiler import __all__ as exposed
__all__ += exposed
import Peeves.Doc as Doc
exposed = [ "Doc" ]
__all__ += exposed



# import sys, os, numpy as np, itertools as ip
import cProfile, pstats, io, abc

__all__ = [
    "BlockProfiler"
]

class AbstractBlockProfiler:
    def __init__(self, name="Profiled Block", inactive=False, print_res=True, logger=None):
        self.name = name
        self.print_res = print_res
        self.inactive = inactive
        self.logger = logger

    @abc.abstractmethod
    def start_profiler(self):
        raise NotImplementedError("abstract class")
    def __enter__(self):
        if not self.inactive:
            self.start_profiler()

    @abc.abstractmethod
    def stop_profiler(self):
        raise NotImplementedError("abstract class")
    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.inactive:
            self.stop_profiler()
            if self.print_res:
                self.print_profile()
    @abc.abstractmethod
    def format_profile(self):
        raise NotImplementedError("abstract class")
    def print_profile(self):
        msg = "In block {}:\n\n{}".format(self.name, self.format_profile())
        if isinstance(self.print_res, str) and self.print_res == 'raise':
            raise ValueError(msg)
        elif self.logger is None:
            if hasattr(self.print_res, 'write'):
                print(msg, file=self.print_res)
            else:
                print(msg)
        else:
            self.logger.log_print(msg)


class PyinstrumentBlockProfiler(AbstractBlockProfiler):

    def __init__(self, name="Profiled Block", inactive=False, print_res=True, logger=None):
        from pyinstrument import Profiler
        super().__init__(name=name, print_res=print_res, inactive=inactive, logger=logger)
        self.profiler = Profiler()

    def start_profiler(self):
        self.profiler.start()

    def stop_profiler(self):
        self.profiler.stop()

    def format_profile(self):
        return self.profiler.output_text(unicode=True, color=True)

class CProfileBlockProfiler(AbstractBlockProfiler):
    """
    Simple class to profile a block of code
    """

    def __init__(self,
                 name="Profile Block",
                 print_res=True,
                 inactive=False,
                 strip_dirs=None,
                 sort_by='cumulative',
                 num_lines=50,
                 filter=None,
                 logger=None
                 ):
        """
        :param name: name of profiled block
        :type name: str
        :param strip_dirs: directory paths to strip from report
        :type strip_dirs: None | Iterable[str]
        """

        super().__init__(name=name, print_res=print_res, inactive=inactive, logger=logger)
        self.strip_dirs = strip_dirs
        self.sort_by = sort_by
        self.num_lines = num_lines
        self.filter = filter
        self.pr = None

    def start_profiler(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def stop_profiler(self):
        self.pr.disable()
        s = io.StringIO()
        sortby = self.sort_by
        ps = pstats.Stats(self.pr, stream=s).sort_stats(sortby)
        filt = [self.num_lines]
        if self.filter is not None:
            if isinstance(self.filter, (int, float, str)):
                filter = [self.filter]
            else:
                filter = list(self.filter)
            filt = filter + filt
        ps.print_stats(*filt)
        self.stat_block = s.getvalue()
        s.close()

    def format_profile(self):
        stat_block = self.stat_block
        strip_dirs = self.strip_dirs
        if strip_dirs is not None:
            for d in self.strip_dirs:
                stat_block = stat_block.replace(d, "")
        return stat_block


def BlockProfiler(name="Profiled Block", print_res=True, mode=None, inactive=False, **kwargs):
    """
    Dispatcher to the various BlockProfiler subclasses
    """

    if mode == None:
        try:
            profiler = PyinstrumentBlockProfiler(name=name, print_res=print_res, inactive=inactive, **kwargs)
        except ImportError:
            profiler = CProfileBlockProfiler(name=name, print_res=print_res, inactive=inactive, **kwargs)
    elif mode == "deterministic":
        profiler = CProfileBlockProfiler(name=name, print_res=print_res, inactive=inactive, **kwargs)
    else:
        profiler = PyinstrumentBlockProfiler(name=name, print_res=print_res, inactive=inactive, **kwargs)


    return profiler


"""
Provides a class that will walk through a set of objects & their children, as loaded into memory, and will generate Markdown for each.
The actual object Markdown is written by the things in the `Writers` module.
"""

import os, types, collections, inspect, importlib, re
from .TemplateEngine import *
from .ObjectWalker import ObjectSpec
from .ExamplesParser import ExamplesParser
from .MarkdownTemplates import MarkdownTemplateFormatter, MarkdownOps

__all__ = [
    "DocWalker",
    "ModuleWriter",
    "ClassWriter",
    "FunctionWriter",
    "MethodWriter",
    "ObjectWriter",
    "IndexWriter"
]

class DocSpec(ObjectSpec):
    """
    A specification for an object to document.
    Supports the fields given by `spec_fields`.
    """
    spec_fields = (
        'id', # name for resolving the object
        'parent', # parent object name for writing,
        'children',
        'examples_root',
        'tests_root'
    )

    def __repr__(self):
        return '{}({})'.format(
            type(self).__name__,
            super().__repr__()
        )

class ExamplesExtractor(TemplateResourceExtractor):
    resource_keys = ['examples']
    resource_attrs = ['__examples__']

class TestsExtractor(TemplateResourceExtractor):
    resource_keys = ['tests']
    resource_attrs = ['__tests__']
    extension = 'Tests.py'
    def path_extension(self, handler:TemplateHandler):
        """
        Provides the default examples path for the object
        :return:
        :rtype:
        """
        base = super().path_extension(handler)
        return [base, os.path.split(base)[1]]
    def load(self, handler:TemplateHandler):
        res = super().load(handler)
        if res is not None:
            res = ExamplesParser(res)
        return res

class TestExamplesFormatter:
    def __init__(self, parser):
        if isinstance(parser, str):
            parser = ExamplesParser(parser)
        self.parser = parser

    @classmethod
    def from_file(cls, tests_file):
        with open(tests_file) as f:
            return cls(f.read())

    def get_template_parameters(self):
        """
        Formats an examples file

        :return:
        :rtype:
        """
        imports = self.parser.setup
        setup = "\n".join(self.parser.format_node(node) for node in imports)

        cls = self.parser.class_spec
        class_setup = "\n".join(
            [
                "class {}({}):".format(
                    cls[0].name,
                    ",".join(x.id for x in cls[0].bases)
                ),
                *(self.parser.format_node(node) for node in cls[1])
            ]
        )

        return dict(
            names=list(self.parser.functions.keys()),
            tests=[self.parser.format_node(v) for v in self.parser.functions.values()],
            setup=setup,
            class_setup=class_setup
        )

class DocTemplateOps(MarkdownOps):
    ...

class DocTemplateHandler(TemplateHandler):
    protected_fields = {'id'}
    default_fields = {'details': "", "related": ""}
    def __init__(self,
                 obj,
                 *,
                 out=None,
                 engine: TemplateEngine=None,
                 root=None,
                 examples_loader:ExamplesExtractor=None,
                 tests_loader:TestsExtractor=None,
                 include_line_numbers=True,
                 walker:'TemplateWalker'=None,
                 **extra_fields
                 ):

        self.examples_loader = examples_loader
        self.tests_loader = tests_loader
        self.include_line_numbers = include_line_numbers
        super().__init__(obj, out=out, engine=engine, root=root, walker=walker, **extra_fields)

    def get_lineno(self):
        # try:
        if self.include_line_numbers:
            obj = self.obj
            if isinstance(obj, (classmethod, staticmethod, property, types.MethodDescriptorType)):
                obj = obj.fget
            elif not isinstance(obj, (type, types.ModuleType, types.FunctionType, types.MethodDescriptorType, types.MethodType)):
                obj = type(obj)
            try:
                lineno = 1 + inspect.findsource(obj)[1] if self.include_line_numbers else ""
            except:
                lineno = ""
        else:
            lineno = ""
        # except:
        #     lineno = ""
        return lineno

    def parse_doc(self, doc):
        """

        :param doc:
        :type doc: str
        :return:
        :rtype:
        """

        # parses a docstring based on reStructured text type specs but Markdown description
        splits = inspect.cleandoc(doc.strip()).splitlines()

        params = collections.deque()
        param_map = {}
        extra_fields = {}

        active_tag = None
        description = []
        for line in splits:
            if line.startswith(":"):
                if line.startswith(":param"):
                    bits = line.split(":", 2)[1:]
                    name = bits[0][5:].strip()
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "Any", "description": []}
                    desc = bits[1].strip() if len(bits) == 2 else ""
                    if len(desc) > 0:
                        param_map[name]["description"].append(desc)
                elif line.startswith(":type"):
                    bits = line.split(":", 2)[1:]
                    name = bits[0][4:].strip()
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "Any", "description": []}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["type"] = t
                elif line.startswith(":return"):
                    bits = line.split(":", 2)[1:]
                    name = ":returns"
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "_", "description": []}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["description"] = t
                elif line.startswith(":rtype"):
                    bits = line.split(":", 2)[1:]
                    name = ":returns"
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "_", "description": []}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["type"] = t
                else:
                    split = line.split()[0]
                    if len(split) > 2 and split.endswith(":"):
                        bits = line.split(":", 2)[1:]
                        name = bits[0]
                        active_tag = name
                        t = bits[1].strip() if len(bits) == 2 else ""
                        if len(t) > 0:
                            extra_fields[name] = [t]
                        else:
                            extra_fields[name] = []
            else:
                if active_tag is None:
                    description.append(line.strip())
                else:
                    if active_tag in param_map:
                        param_map[active_tag]['description'].append(line)
                    else:
                        extra_fields[active_tag].append(line)
        for v in param_map.values():
            v['description'] = "\n".join(v['description'])
        for k, v in extra_fields.items():
            extra_fields[k] = "\n".join(v)

        return inspect.cleandoc("\n".join(description)).strip(), param_map, extra_fields

    def load_examples(self):
        return self.examples_loader.load(self) if self.examples_loader is not None else None
    def load_tests(self):
        tests = self.tests_loader.load(self) if self.tests_loader is not None else None
        if tests is None:
            parent_tests = self.resolve_key('parent_tests')
            if parent_tests is not None:
                tests = parent_tests.filter_by_name(self.name.split(".")[-1])
        if tests is not None:
            self.extra_fields['parent_tests'] = tests
            tests = TestExamplesFormatter(tests).get_template_parameters()
        return tests

class DocObjectTemplateHandler(DocTemplateHandler):
    def get_package_and_url(self, include_url_base=True):
        pkg, file_url=super().get_package_and_url(include_url_base=include_url_base)
        try:
            base_url, ext = file_url.rsplit("/", 1)
        except ValueError:
            pass
        else:
            if ext == '__init__.py':
                try:
                    base_url, ext = base_url.rsplit("/", 1)
                except ValueError:
                    pass
            file_url = base_url + '.py'
        return pkg, file_url

    def load_examples(self):
        return self.examples_loader.load(self) if self.examples_loader is not None else None
    def load_tests(self):
        tests = self.tests_loader.load(self) if self.tests_loader is not None else None
        if tests is None:
            parent_tests = self.resolve_key('parent_tests')
            if parent_tests is not None:
                tests = parent_tests.filter_by_name(self.name.split(".")[-1])
        if tests is not None:
            self.extra_fields['parent_tests'] = tests
            tests = TestExamplesFormatter(tests).get_template_parameters()
        return tests
class ModuleWriter(DocTemplateHandler):
    """A writer targeted to a module object. Just needs to write the Module metadata."""
    template = 'module.md'
    def __init__(self, obj, **kwargs):
        if isinstance(obj, str):
            obj = importlib.import_module(obj)
        super().__init__(obj, **kwargs)

    def get_template_params(self):
        """
        Provides module specific parameters
        :return:
        :rtype:
        """

        mod = self.obj  # type: types.ModuleType
        name = mod.__name__
        ident = self.identifier
        ident_depth = len(ident.split("."))
        # get identifiers
        idents = [".".join(self.get_target_extension(getattr(mod, a))) for a in self.get_members(mod)]
        # flattend them
        idents = [i for i in idents if ident in i]
        # split by qualified names
        idents = [".".join(a.split(".")[ident_depth-1:]) for a in idents]
        descr, _, fields = self.parse_doc(mod.__doc__ if mod.__doc__ is not None else '')
        # long_descr = mod.__long_doc__ if hasattr(mod, '__long_doc__') and mod.__long_doc__ is not None else ''

        ex = self.load_examples()
        tests = self.load_tests()
        return dict({
            'id': ident,
            'description': descr.strip(),
            # 'long_description': long_descr.strip(),
            'name': name,
            'members': idents,
            'examples': ex,
            'tests': tests,
            'lineno' : self.get_lineno()
        }, **fields)

    @classmethod
    def get_members(cls, mod):
        return (mod.__all__ if hasattr(mod, '__all__') else [])

class ClassWriter(DocObjectTemplateHandler):
    """A writer targeted to a class"""

    template = 'class.md'
    def load_methods(self, function_writer=None):
        """
        Loads the methods supported by the class

        :param function_writer:
        :type function_writer:
        :return:
        :rtype:
        """

        if function_writer is None:
            function_writer = self.walker.method_handler

        cls = self.obj
        keys = cls.__all__ if hasattr(cls, '__all__') else list(cls.__dict__.keys())

        props = []
        methods = []

        extra_fields = self.extra_fields.copy()
        pkg, file_url = self.package_path
        extra_fields['package_name'] = pkg
        extra_fields['file_url'] = file_url
        extra_fields['package_url'] = os.path.dirname(file_url)
        for k in keys:
            o = getattr(cls, k)
            if isinstance(o, (types.FunctionType, types.MethodType, classmethod, staticmethod, property)):
                if not k.startswith("_") or (k.startswith("__") and k.endswith("__")):
                    methods.append(
                        self.walker.get_handler(
                            o,
                            cls=function_writer,
                            name=k,
                            out_file=None,
                            extra_fields=extra_fields,
                            parent=self.identifier
                        )
                    )
            else:
                if not k.startswith("_"):
                    props.append((k, o))

        return props, methods

    def format_prop(self, k, o):
        return '{}: {}'.format(k, type(o).__name__)

    def get_template_params(self, function_writer=None):
        """

        :param function_writer:
        :type function_writer:
        :return:
        :rtype:
        """

        cls = self.obj  # type: type
        ex = self.load_examples()
        tests = self.load_tests()
        name = cls.__name__
        ident = self.identifier
        props, methods = self.load_methods(function_writer=function_writer)
        descr, param, fields = self.parse_doc(cls.__doc__ if cls.__doc__ is not None else '')
        lineno = self.get_lineno()

        return dict({
            'id': ident,
            'name': name,
            'lineno': lineno,
            'description': descr,
            'parameters': param,
            'props': props,
            'methods': methods,
            'examples': ex if ex is not None else "",
            'tests': tests
        }, **fields)

class FunctionWriter(DocObjectTemplateHandler):
    """
    Writer to dump functions to file
    """

    template = 'function.md'

    def get_signature(self):
        return str(inspect.signature(self.obj))

    def get_template_params(self, **kwargs):
        f = self.obj  # type: types.FunctionType
        ident = self.identifier
        signature = self.get_signature()
        mem_obj_pat = re.compile(" object at \w+>")
        signature = re.sub(mem_obj_pat, " instance>", signature)
        name = self.get_name()
        descr, param, fields = self.parse_doc(f.__doc__ if f.__doc__ is not None else '')
        ex = self.load_examples()
        tests = self.load_tests()
        lineno = self.get_lineno()
        return dict({
            'id': ident,
            'name': name,
            'lineno': lineno,
            'signature': signature,
            'parameters': param,
            'description': descr,
            'examples': ex if ex is not None else "",
            'tests': tests
        }, **fields)

class MethodWriter(FunctionWriter):
    """
    Writes class methods to file
    (distinct from functions since not expected to exist solo)
    """

    template = 'method.md'

    def get_template_params(self, **kwargs):
        params = super().get_template_params(**kwargs)
        meth = self.obj  # type: types.MethodType
        decorator = ""
        if isinstance(meth, classmethod):
            decorator = 'classmethod'
        elif isinstance(meth, property):
            decorator = 'property'
        elif isinstance(meth, staticmethod):
            decorator = 'staticmethod'
        if len(decorator) > 0:
            decorator = "@" + decorator + "\n"
        params['decorator'] = decorator
        return params

    def get_signature(self):
        try:
            signature = str(inspect.signature(self.obj))
        except TypeError:  # dies on properties
            signature = "(self)"
        return signature

    @property
    def identifier(self):
        if isinstance(self.obj, property):
            return self.get_identifier(self.resolve_parent(check_tree=False)) + "." + self.get_name()
        else:
            return self.get_identifier(self.obj)

class ObjectWriter(DocObjectTemplateHandler):
    """
    Writes general objects to file.
    Basically a fallback to support singletons and things
    of that nature.
    """

    template = 'object.md'

    @property
    def identifier(self):
        try:
            qualname = self.obj.__qualname__
        except AttributeError:
            qualname = self.get_identifier(type(self.obj)) + "." + self.get_name()
        qn = qualname.split(".")
        qualname = ".".join(qn[:-2] + qn[-1:])  # want to drop the class name
        return qualname

    def check_should_write(self):
        """
        Determines whether the object really actually should be
        documented (quite permissive)
        :return:
        :rtype:
        """
        return (
                hasattr(self.obj, "__doc__")
                and hasattr(self.obj, "__name__")
                and super().check_should_write()
        )

    def get_template_params(self):

        try:
            doc = self.obj.__doc__
        except AttributeError:
            descr = "instance of " + type(self.obj).__name__
            fields = {}
        else:
            descr, _, fields = self.parse_doc(doc if doc is not None else '')

        if descr is None:
            descr = ''

        ex = self.load_examples()
        lineno = self.get_lineno()
        return dict({
            'id': self.identifier,
            'lineno': lineno,
            'name': self.get_name(),
            'description': descr.strip(),
            'examples': ex if ex is not None else "",
            'tests': None
        }, **fields)

class IndexWriter(DocTemplateHandler):
    """
    Writes an index file with all of the
    written documentation files.
    Needs some work to provide more useful info by default.
    """
    template = 'index.md'

    def __init__(self, *args, description=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.description = description if description is not None else "# Documentation"

    def get_identifier(cls, o):
        return 'index'

    def get_file_paths(self):
        rl = len(os.path.split(self.root))
        fs = ["/".join(os.path.split(f)[rl - 1:]) for f in self.obj]
        return fs

    # def format_item(self, f):
    #     self.formatter.format_item(
    #         self.formatter.format_link(os.path.splitext(f.split("/")[-1])[0], f)
    #     )

    def get_template_params(self):
        descr, _, fields = self.parse_doc(self.description if self.description is not None else '')
        return dict({
            'index_files': [[os.path.splitext(os.path.basename(f))[0], f] for f in self.get_file_paths()],
            'description': descr,
            'examples': self.load_examples()
        }, **fields)

class DocWalker(TemplateWalker):
    """
    A class that walks a module structure, generating `.md` files for every class inside it as well as for global functions,
    and a Markdown index file.

    Takes a set of objects & writers and walks through the objects, generating files on the way.

    :details: A `DocWalker` object is a light subclass of a `TemplateWalker`, but specialized for documentation & with specialized handlers
    :related: .DocsBuilder.DocBuilder, ModuleWriter, ClassWriter, FunctionWriter, MethodWriter, ObjectWriter, IndexWriter
    """

    module_handler = ModuleWriter
    class_handler = ClassWriter
    function_handler = FunctionWriter
    method_handler = MethodWriter
    object_handler = ObjectWriter
    index_handler = IndexWriter
    spec = DocSpec
    def __init__(self,
                 out=None,
                 engine=None,
                 verbose=True,
                 template_locator=None,
                 examples_directory=None,
                 tests_directory=None,
                 **extra_fields
                 ):
        """
        :param objects: the objects to write out
        :type objects: Iterable[Any]
        :param out: the directory in which to write the files (`None` means `sys.stdout`)
        :type out: None | str
        :param out: the directory in which to write the files (`None` means `sys.stdout`)
        :type out: None | str
        :param: writers
        :type: DispatchTable
        :param ignore_paths: a set of paths not to write (passed to the objects)
        :type ignore_paths: None | Iterable[str]
        """

        if engine is None:
            engine = self.get_engine(template_locator)
        self.examples_loader = self.get_examples_loader(examples_directory)
        self.tests_loader = self.get_tests_loader(tests_directory)
        self.verbose = verbose

        super().__init__(engine, out=out, **extra_fields)

    def get_engine(self, locator):
        if not isinstance(locator, TemplateEngine):
            locator = TemplateEngine(locator, template_pattern="*.md", formatter_class=MarkdownTemplateFormatter)
        return locator
    def get_examples_loader(self, examples_directory):
        examples_loader = examples_directory
        if examples_loader is not None and not isinstance(examples_loader, ExamplesExtractor):
            examples_loader = ExamplesExtractor(examples_loader)
        return examples_loader
    def get_tests_loader(self, tests_directory):
        tests_loader = tests_directory
        if tests_loader is not None and not isinstance(tests_loader, TestsExtractor):
            tests_loader = TestsExtractor(tests_loader)
        return tests_loader
    def get_handler(self, *args, examples_loader=None, tests_loader=None, **kwargs):
        return super().get_handler(
            *args,
            examples_loader=self.examples_loader if examples_loader is None else examples_loader,
            tests_loader=self.tests_loader if tests_loader is None else tests_loader,
            **kwargs
        )

    def visit_root(self, o, tests_directory=None, examples_directory=None, **kwargs):
        if tests_directory is None and isinstance(o, dict):
            tests_directory = o.get('tests_directory', None)
        old_tl = self.tests_loader
        if tests_directory is not None:
            self.tests_loader = self.get_tests_loader(tests_directory)
        if examples_directory is None and isinstance(o, dict):
            examples_directory = o.get('examples_directory', None)
        old_el = self.examples_loader
        if examples_directory is not None:
            self.examples_loader = self.get_examples_loader(examples_directory)
        try:
            return super().visit_root(o, **kwargs)
        finally:
            self.tests_loader = old_tl
            self.examples_loader = old_el





"""
Peeves
A simple package for doing docs and testing
"""

# pulled from Ryan's stuff, lightly modified
from setuptools import setup, find_packages

short_description = __doc__.split("\n")

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except:
    long_description = "\n".join(short_description[2:])

def get_version():
    import subprocess
    run_out = subprocess.run(['git', 'describe', '--tags', '--abbrev=0'], capture_output=True)
    return run_out.stdout.decode().strip().strip("v")

setup(
    # Self-descriptive entries which should always be present
    name='mccoygroup-peeves',
    author='Mark Boyer',
    author_email='b3m2a1@uw.edu',
    description=short_description[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=get_version(),
    license='MIT',

    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages()

    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    # include_package_data=True

)



import re, uuid, pathlib
from .TemplateEngine import *
from .HTML import *

class MarkdownOps:
    @classmethod
    def format_item(self, item, item_level = 0):
        return "{}- {}".format('  ' * (item_level + 1), item)
    @classmethod
    def format_link(self, alt, link):
        return '[{}]({})'.format(alt, link)
    @classmethod
    def format_obj_link(self, spec):
        return self.format_link(self.canonical_name(spec), self.canonical_link(spec))
    @classmethod
    def format_inline_code(self, arg):
        """

        :param arg:
        :type arg: str
        :return:
        :rtype:
        """
        nticks = arg.count("`")
        fence = "`"*(nticks+1)
        return fence + arg + fence
    @classmethod
    def format_code_block(self, arg):
        """

        :param arg:
        :type arg: str
        :return:
        :rtype:
        """
        nticks = arg.count("`")
        fence = "`"*(nticks+3)
        return fence + "python\n" + arg + "\n" + fence
    @classmethod
    def format_quote_block(self, arg):
        """

        :param arg:
        :type arg: str
        :return:
        :rtype:
        """

        return ">" + arg.replace("\n", "\n>")

    link_bar_template='<div class="container{boxed}">\n{links}\n</div>'
    link_row_template='  <div class="row">\n{cols}\n</div>'
    link_item_template='   <div class="col" markdown="1">\n{item}   \n</div>'
    @classmethod
    def format_grid(self, link_grid, boxed=False):
        return self.link_bar_template.format(links="\n".join(
            self.link_row_template.format(
                cols="\n".join(self.link_item_template.format(item=item) for item in row)
            )
            for row in link_grid if len(row) > 0
        ),
        boxed=' alert alert-secondary bg-light' if boxed else ""
        )
    @classmethod
    def split(self, links, ncols=3, pad=""):
        num_cols = ncols
        splits = []
        sub = []
        for x in links:
            sub.append(x)
            if len(sub) == num_cols:
                splits.append(sub)
                sub = []
        splits.append(sub + [pad] * (ncols - len(sub)))
        return splits

    collapse_template="""
<div class="collapsible-section">
 <div class="collapsible-section collapsible-section-header" markdown="1">
{header_fmt} <a class="collapse-link" data-toggle="collapse" href="#{name}" markdown="1">{header}</a> {opener}
 </div>
 <div class="collapsible-section collapsible-section-body collapse {show}" id="{name}" markdown="1">
 {content}
 </div>
</div>
"""
    collapse_opener = '<a class="float-right" data-toggle="collapse" href="#{name}"><i class="fa fa-chevron-down"></i></a>'
    @classmethod
    def format_collapse_section(self, header, content, name=None, open=True, include_opener=True):
        header_fmt = ""
        while header.startswith("#"):
            header_fmt += "#"
            header = header[1:]
        if name is None:
            name = re.sub("\W", "", header) + "-" + str(uuid.uuid4())[:6]
        return self.collapse_template.format(
            header_fmt=header_fmt,
            header=header,
            content=content,
            name=name,
            show="show" if open else "",
            opener=self.collapse_opener.format(name=name) if include_opener else ""
        )

    @classmethod
    def format_obj_link_grid(self, mems, ncols=3, boxed=True):
        links = self.split(
            [self.format_obj_link(l) for l in mems],
            ncols=ncols
        )
        return self.format_grid(links, boxed=boxed)

    @classmethod
    def canonical_name(self, identifier, formatter=None):
        return identifier.split(".")[-1]

    @classmethod
    def canonical_link(self, identifier, formatter=None):
        ups = 0
        while identifier[0] == ".":
            ups += 1
            identifier = identifier[1:]
        if ups > 0:
            pad = "../"*ups
        else:
            pad = ""
        identifier = "/".join(identifier.split("."))
        return pad + identifier + ".md"

    @classmethod
    def html(kls, tag, content, markdown=True, formatter=None, **styles):
        if markdown:
            styles["markdown"] = "1"
        return getattr(HTML, tag)("\n{content}\n", **styles).tostring().format(content=content)
    @classmethod
    def bootstrap(kls, tag, content, markdown=True, formatter=None, **styles):
        if markdown:
            styles["markdown"] = "1"
        return getattr(Bootstrap, tag)("\n{content}\n", **styles).tostring().format(content=content)
    @classmethod
    def alert(kls, content, variant='warning', markdown=True, formatter=None, **styles):
        return kls.bootstrap('Alert', content, variant=variant, markdown=markdown, **styles)

class MarkdownFormatDirective(FormatDirective):
    Link = "link", TemplateOps.wrap(MarkdownOps.format_link)
    ObjLink = "objlink", TemplateOps.wrap(MarkdownOps.format_obj_link)
    Item = "item", TemplateOps.wrap(MarkdownOps.format_item)
    Code = "code", TemplateOps.wrap(MarkdownOps.format_code_block)
    Quote = "quote", TemplateOps.wrap(MarkdownOps.format_quote_block)
    # Card = "card", TemplateOps.wrap(MarkdownFormatter.format_card)
    # Alert = "alert", TemplateOps.wrap(MarkdownFormatter.format_alert)
    Collapse = "collapse", TemplateOps.wrap(MarkdownOps.format_collapse_section)
    Grid = "grid", TemplateOps.wrap(MarkdownOps.format_grid)
    Split = "split", TemplateOps.wrap(MarkdownOps.split)
    ObjLinkGrid = "objlink_grid", TemplateOps.wrap(MarkdownOps.format_obj_link_grid)
    CanonicalName = "canonical_name", MarkdownOps.canonical_name
    CanonicalLink = "canonical_link", MarkdownOps.canonical_link
    HTML = "html", MarkdownOps.html
    Bootstrap = "bootstrap", MarkdownOps.bootstrap
    Alert = "alert", MarkdownOps.alert
MarkdownFormatDirective = TemplateFormatDirective.extend(MarkdownFormatDirective)

class MarkdownTemplateFormatter(TemplateFormatter):
    directives = MarkdownFormatDirective
